// Graph.cpp: implementation of the Graph class.
//
//     
//
//
// *****     MFC operation+= CRect
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Color.h"
#include "Graph.h"
#include "Nucleus.h"
#include "math.h"


#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

Graph::Graph()
{
	Graph_();
	linear[1][0]=linear[0][1]=1;
	linear[0][1]=linear[1][0]=0;
}

//   

void Graph::Graph_()
{
	POSITION.angle=0;
	POSITION.scaly=1;
	POSITION.offset.x=POSITION.offset.y=0;
	SIZES.isCenter=SIZES.isSize=false;
}

Graph::~Graph()
{
}

//       

int Graph::FindCurrent(CPoint point,int r,int skip)
{
	int min=100,num;
	for(int n=0; n<tops.GetSize(); n++) { 
		if(n==skip) continue;
		int d=GetDistance(n,point);
		if (d<min) min=d,num=n;
	}
	if (min>r) return -1;
	UnselectAll();
	return num;
}

//   

bool Graph::Add(CPoint point)
{
	if (point.x<TopSize || point.y<TopSize) return false;
	Canonize(); point-=POSITION.offset;	if (SIZES.isSize) UpdateSize(point);
	topST top; memset(&top,0,sizeof(top)); top.point=point;
	if (FindCurrent(point,TopSize*2)<0) tops.Add(top); else return false;
	UnselectAll();
	return true;
}

//   

bool Graph::AddLink(int x, int y,int curve)
{
	CPoint point;
	if(y<x) point=CPoint(x,y); else point=CPoint(y,x); 
	int n=GetLINKs(); while(n) { if(point==links[--n].point) return true;}
	linkST link; memset(&link,0,sizeof(link));
	link.point=point; link.curve=curve; links.Add(link);
	Analiz();
	return false;
}

//   

void Graph::Set(int index,CPoint point)
{
	if (point.x<TopSize || point.y<TopSize) return;
	Canonize(); if (SIZES.isSize)
	{
		CPoint p=GetTOP(index);
		if((p.x==SIZES.rect.left || p.x==SIZES.rect.right || p.y==SIZES.rect.top || p.y==SIZES.rect.bottom)
		&&(point.x>SIZES.rect.left && point.x<SIZES.rect.right || point.y>SIZES.rect.bottom && point.y<SIZES.rect.top)) SIZES.isSize=false;
		else UpdateSize(point);
	}
	point-=POSITION.offset;
	if (FindCurrent(point,TopSize*2,index)<0) tops[index].point=point;
}

//  

void Graph::Del(int index)
{
	CPoint point=GetTOP(index);
	if(point.x==SIZES.rect.left || point.x==SIZES.rect.right || point.y==SIZES.rect.top || point.y==SIZES.rect.bottom) SIZES.isSize=false;
	for(int n=0; n<GetLINKs();)
	{
		point=links[n].point;
		if(point.x==index || point.y==index) { DelLink(n); continue;}
		if(point.x>index) point.x--;
		if(point.y>index) point.y--;
		links[n++].point=point;
	}
	tops.RemoveAt(index);
}

void Graph::DelLink(int index)
{
	links.RemoveAt(index);
}

//     

int Graph::GetDistance(int index, CPoint point)
{
	return Distance(point,GetTOP(index));
}

//    

int Graph::Distance(CPoint point1, CPoint point2)
{
	point1-=point2;
	return (int)sqrt(point1.x*point1.x+point1.y*point1.y);
}

//     

CPoint Graph::GetCircleCenter(CPoint point1, CPoint point2, int r)
{
	double tn=tan(acos((double)Distance(point1,point2)/2/r))/2;
//	int d=Distance(point1,point2); double tn=sqrt(r*r-d*d/4)/d;
	return CPoint((int)((point1.x+point2.x)/2-(point2.y-point1.y)*tn),(int)((point1.y+point2.y)/2+(point2.x-point1.x)*tn));
} 

//  

int Graph::GetCurve(int index)
{
	return links[index].curve;
}

//        (  )

bool Graph::HitToLink(int index, CPoint point)
{
	CPoint point1=GetTOP(links[index].point.x);
	CPoint point2=GetTOP(links[index].point.y);
	int n=Distance(point1,point2); int r=(int)(links[index].curve)*(n/2+5)/65536;
	if (r) return abs(Distance(point,GetCircleCenter(point1,point2,r))-abs(r))<5;
	return n+5>Distance(point,point1)+Distance(point,point2);
}

//    

bool Graph::isSelect(int index)
{
	return links[index].select;
}

//      ...     .

bool Graph::SelectLink(CPoint point)
{
	bool res = false;
	for(int n=0; n<links.GetSize(); n++) 
	{
		if (isSelect(n) != HitToLink(n,point)) links[n].select=!links[n].select, res=true;
	}
	return res;
}

//        

bool Graph::SolveCurve(CPoint point)
{
	bool res=false;  int a,b,c,r;
	CPoint point1,point2,point3,point4;
	for(int n=0; n<GetLINKs(); n++) {
		if (!isSelect(n)) continue;
			point4=links[n].point;
			point1=GetTOP(point4.x);
			point2=GetTOP(point4.y);
			point3.x=(point1.x+point2.x)/2;
			point3.y=(point1.y+point2.y)/2;
			a=Distance(point1,point2); 
			b=Distance(point,point1); 
			c=Distance(point,point2);
			if(Distance(point,point3)+25>a/2) r=a/2+5; 
				else r=(int)(a/2/sin(acos(((double)(b*b+c*c-a*a))/2/b/c)));
			if(abs(r)>2000) r=0;
			point1-=point2; point2-=point;
			if((point1.x)*(point2.y)>(point1.y)*(point2.x)) r=-r;
			r=r*65536/(a/2+5);
			res=res || ((int)(links[n].curve)!=r);
			links[n].curve=r;
	}
	return res;
}


CPoint Graph::GetTOP(int index)
{
	CPoint point;
	point=tops[index].point;
	if(POSITION.scaly!=1 || POSITION.angle) {
		point-=SIZES.center;
		point=CPoint((int)(linear[0][0]*point.x+linear[0][1]*point.y),(int)(linear[1][0]*point.x+linear[1][1]*point.y))+SIZES.center;
	}
	return point+POSITION.offset;
}


int Graph::GetTOPs()
{
	return tops.GetSize();
}

void Graph::SplitLink(int index, int n)
{
	CPoint point=links[index].point;
	int curves=GetCurve(index);
	DelLink(index);
	AddLink(point.x,n,curves);
	AddLink(point.y,n,curves);
}

int Graph::GetLINKs()
{
	return links.GetSize();
}

void Graph::Analiz()
{
	Nucleus nucleus;
	nucleus.Start(this);
}

int Graph::GetColor(int index)
{
	const DWORD ink[]={
		RGB(255,0,0),     //0  light red
		RGB(0,255,0),     //1  light green
		RGB(0,0,255),     //2  light blye
		RGB(255,255,0),   //3  light yellow
		RGB(0,255,255),   //4  light cyan
		RGB(255,0,255),   //5  light magenta
		RGB(192,192,192), //6  light gray
		RGB(128,0,0),     //7  dark red
		RGB(0,128,0),     //8  dark green
		RGB(0,0,128),     //9  dark blye
		RGB(128,128,0),   //10 dark yellow
		RGB(0,112,128),   //11 dark cyan
		RGB(128,0,128),   //12 dark magenta
		RGB(128,128,128), //13 dark gray
	};
	if(index<0 || index>13) return 0; 
	return ink[links[index].color];
}

//    

void Graph::EvalueSize()
{
	if(SIZES.isSize) return;
	CPoint point; CRect rect;
	if(GetTOPs()) point=GetTOP(0); else point.x=point.y=0;
	SIZES.rect=CRect(point,point);
	for(int m=0; m<GetTOPs(); m++) UpdateSize(GetTOP(m));
	SIZES.isSize=true;
}

//      

void Graph::UpdateSize(CPoint point)
{
	if (!GetTOPs()) { SIZES.rect=CRect(point,point); return; }
	if (point.x<SIZES.rect.left)   SIZES.rect.left  =point.x;
	if (point.x>SIZES.rect.right)  SIZES.rect.right =point.x;
	if (point.y<SIZES.rect.top)    SIZES.rect.top   =point.y;
	if (point.y>SIZES.rect.bottom) SIZES.rect.bottom=point.y;
}


//      ,   

void Graph::Canonize()
{
	if(POSITION.angle || POSITION.scaly!=1) {
		for(int n=0; n<GetTOPs(); n++) tops[n].point=GetTOP(n);
		Graph_();
	}
}

//   

void Graph::EvalueCenter()
{
	if(SIZES.isCenter) return;
	EvalueSize();
	SIZES.center=CPoint((SIZES.rect.left+SIZES.rect.right)/2,(SIZES.rect.top+SIZES.rect.bottom)/2)-POSITION.offset;
	SIZES.isCenter=true;
}

//  

int Graph::Rotate(int angle)
{
	if (angle!=GET_ANGLE) {
		SIZES.isSize=false; EvalueCenter();
		POSITION.angle=angle;
		while(POSITION.angle<0) POSITION.angle+=360;
		while(POSITION.angle>360) POSITION.angle-=360;
		MakeLinear();
	}
	return POSITION.angle;
}

//  
void Graph::Move(CPoint point)
{
	SIZES.rect+=CRect(-point,point);
	POSITION.offset+=point;
	EvalueSize();
	//return POSITION.offset;
}

//   

double Graph::Scaly(double x)
{
	EvalueCenter();
	if(x) {
		int r_max=99999,quantity=GetTOPs();
		for(int index1=0; index1<quantity; index1++) {
			CPoint point=tops[index1].point;
			for(int index2=index1+1; index2<quantity; index2++) {
				int r=Distance(point,GetTOP(index2));
				if(r<r_max) r_max=r;
			}
		}
		POSITION.scaly=x;
		if(r_max*POSITION.scaly<TopSize*2.4) POSITION.scaly=(double)(TopSize*2.4)/r_max;
		SIZES.isSize=false;
		MakeLinear();
	}
	return POSITION.scaly;
}

//    

CRect Graph::GetSize()
{
	EvalueSize(); CPoint point(0,0);
	if (GetTOPs()) {
		if(SIZES.rect.left<TopSize) point.x=TopSize-SIZES.rect.left;
		if(SIZES.rect.top<TopSize) point.y=TopSize-SIZES.rect.top;
		POSITION.offset+=point;
		SIZES.rect+=CRect(-point,point);
	}
	return SIZES.rect;
}

//         .....

void Graph::MakeLinear()
{
	double f=2*PI*POSITION.angle/360;
	linear[0][0]=+POSITION.scaly*cos(f); linear[0][1]=-POSITION.scaly*sin(f);
	linear[1][0]=+POSITION.scaly*sin(f); linear[1][1]=+POSITION.scaly*cos(f);
}

CPoint Graph::GetCenter()
{
	EvalueCenter();
	return SIZES.center+POSITION.offset;
}

CPoint Graph::GetLINK(int index)
{
	return links.GetAt(index).point;
}

int Graph::GetZ(int index)
{
	return tops[index].z;
}

void Graph::SetZ(int index, int z)
{
	tops[index].z=z;
}

void Graph::ReportClear()
{
//	ReportDialog dlg;
//	dlg.DoModal();
}

void Graph::UnselectAll()
{
	int n=GetLINKs();
	while(n) links[--n].select=false;
}

void Graph::SetColor(WORD index,WORD color)
{
	links[index].color=color;
}

void Graph::Save(FILE *handler)
{
	int n; Canonize();
	n=GetTOPs(); fwrite(&n,1,sizeof(n),handler);
	for(n=0; n<GetTOPs(); n++) fwrite(&tops[n],1,sizeof(topST),handler);
	n=GetLINKs(); fwrite(&n,1,sizeof(n),handler);
	for(n=0; n<GetLINKs(); n++) fwrite(&links[n],1,sizeof(linkST),handler);
}

void Graph::Load(FILE *handler)
{
	int n; linkST b; topST a;
	tops.RemoveAll(); links.RemoveAll();
	fread(&n,1,sizeof(n),handler);
	while(n--) fread(&a,1,sizeof(topST),handler),tops.Add(a);
	fread(&n,1,sizeof(n),handler);
	while(n--) fread(&b,1,sizeof(linkST),handler),links.Add(b);
}

#include "TEXsetup.h"

void Graph::SaveAsTEX(FILE *handler)
{
	TEXsetup dlg; int s1,s2; int n; CPoint point,point1,point2;
	dlg.DoModal();
	s1=dlg.scaly1; s2=dlg.scaly2;
	fprintf(handler,"\\begin{picture}(%d.00,%d.00)",SIZES.rect.right,SIZES.rect.bottom);
	for(n=0; n<GetLINKs(); n++)
	{
		point =GetLINK(n);
		point1=GetTOP(point.x);
		point2=GetTOP(point.y);
		fprintf(handler,"\\emline{%d.00}{%d.00}{%d}{%d.00}{%d.00}{%d}\n",point1.x,point1.y,n+n, point2.x,point2.y,n+n+1);
	}

	for(n=0; n<GetTOPs(); n++)
	{
		point=GetTOP(n);
		fprintf(handler,"\\put(%d.00,%d.00){\\circle*{4}}\n", point.x,point.y);
		fprintf(handler,"\\put(%d.00,%d.00){\\makebox(0,0)[cc]{%d}}\n",point.x+10,point.y-10,n);
	}
	fprintf(handler,"\\end{picture}");
//http://www.jeenn.dp.ua/ru/worx/basis.htm
}

void Graph::SaveAsMatrix(FILE *handler)
{
	WORD x,y,size=GetTOPs(); int n; CPoint point;
	ContiguityMatrix matrix(size);
	for(n=0; n<GetLINKs(); n++) matrix.Set(GetLINK(n),true);
	for(y=0; y<size; y++) {for(x=0; x<size; x++) 
		putc(' ',handler); putc(matrix.Get(x,y) ? '1' : '0',handler);
	putc('\n',handler); }
	putc('\n',handler);
	for(n=0; n<GetLINKs(); n++) point=GetLINK(n),fprintf(handler,"%d %d %d\n",point.x,point.y,GetZ(n));
}
