#include "stdafx.h"
#include "SD.h"

struct node{node *next1;node *next2;WORD x;WORD y;}*tree;

int IndexByMatrix(WORD x,WORD y)
{
	return (y-1)*y/2+x;
}

void RemoveTree(void *tree)
{
	if(tree) RemoveTree(((node*)tree)->next1),
		RemoveTree(((node*)tree)->next2),
		delete[](tree);
}

bool CheckIndex(WORD x,WORD y,WORD size)
{
	if(x<size && y<size) return false;
	CString text; text.Format("Invalid index [%d %d %d]",(int)x,(int)y,(int)size);
	MessageBox(0,text,"Matrix",0);
	return true;
}


//////////////////////////////////////////////////////////////////////////////////

BitMap::BitMap(WORD n)
{
	size=n;
	array=new(BYTE[size>>3+1]);
}

BitMap::~BitMap()
{
	delete[]array;
}

void BitMap::Clear()
{
	memset(array,0,size>>3+1);
}

void BitMap::Fill()
{
	memset(array,-1,size>>3+1);
}

bool BitMap::Get(WORD index)
{
	return ((array[index>>3])&(1<<(index&7)))!=0;
}

void BitMap::Set(WORD index)
{
	Set(index,true);
}

void BitMap::Res(WORD index)
{
	Set(index,false);
}

void BitMap::Set(WORD index, bool to)
{
	array[index>>3]|=(1<<(index&7));
	if(!to) array[index>>3]^=(1<<(index&7));
}

WORD BitMap::GetSize()
{
	return size;
}

//////////////////////////////////////////////////////////////////////////////////

ContiguityMatrix::ContiguityMatrix()
{
	tree=0;
	matrix=0;
}

ContiguityMatrix::ContiguityMatrix(WORD size,bool n)
{
	matrix=0; tree=0;
	Install(size,n);
}

ContiguityMatrix::~ContiguityMatrix()
{
	Uninstall();
}

void ContiguityMatrix::Install(WORD size_,bool n) 
{
	size=size_; 
	nodeLimit=size*(size-1)/8/sizeof(*tree)+2;
        Uninstall(); def=n;
}

void ContiguityMatrix::Uninstall()
{
	if(matrix) delete[]matrix; matrix=0;
	if(tree) RemoveTree(tree); tree=0;
}

//----------------------------------------------------------------

int ContiguityMatrix::GetSize()
{
	return size;
}

void ContiguityMatrix::Set(CPoint point,bool n)
{
	Set((WORD)point.x,(WORD)point.y,n);
}

void ContiguityMatrix::Set(WORD x,WORD y,bool v)
{
	if(CheckIndex(x,y,size)) return;
	if(x==y) return; WORD t; if(y<x) t=x,x=y,y=t;
	if (matrix) SetByMatrix(x,y,v); else SetByTree(x,y,v);
}

bool ContiguityMatrix::Get(WORD x, WORD y)
{
	if(CheckIndex(x,y,size)) return def;
	if(x==y) return def; WORD t; if(y<x) t=x,x=y,y=t;
	return (matrix ? GetByMatrix(x,y) : GetByTree(x,y));
}

//----------------------------------------------------------------
// Matrix as Triangle (c) Proteus

bool ContiguityMatrix::GetByMatrix(WORD x,WORD y)
{
	int n=IndexByMatrix(x,y);
	return (matrix[n>>3] & (1<<(n&7)))!=0;
}

void ContiguityMatrix::SetByMatrix(WORD x,WORD y,bool v)
{
	int n=IndexByMatrix(x,y),m=1<<(n&7);
	matrix[n>>3]|=m;
	if(!v) matrix[n>>3]^=m;
}

void ContiguityMatrix::MakeMatrix()
{
	if(size>1) matrix=new(BYTE[size*(size-1)/2/8+1]),memset(matrix,def,size*(size-1)/2/8+1);
}

//----------------------------------------------------------------
// Matrix as BSP tree  (c) Proteus

void *ContiguityMatrix::FindByTree(WORD x,WORD y,bool add)
{
	node *next,*root=tree; bool direction,split=false;
	WORD s,sx,sy; s=sx=sy=size/2;
	for(;;)
	{
		if(root->x==x && root->y==y) return root;
		direction=(split?(direction=root->x>sx):(direction=root->y>sy));
		s/=2;
		if(split) {	if(direction) sx+=s; else sx-=s;} else {if(direction) sy+=s; else sy-=s;}
		next=(direction ? root->next1 : root->next2);
		if(next) root=next; else {
			if(!add) return 0; next=new(node); next->next1=next->next2=0; next->x=x; next->y=y; nodeCount++;
			if(direction) root->next1=next; else root->next2=next;
			return next;
		}
	}
}


void ContiguityMatrix::TreeToMatrix(node *tree)
{
	if(tree)
	{
		SetByMatrix(tree->x,tree->y,tree->value);
		TreeToMatrix(tree->next1);
		TreeToMatrix(tree->next2);
	}
}

bool ContiguityMatrix::GetByTree(WORD x,WORD y)
{
	if(!tree) return def;
	node *index=(node*)FindByTree(x,y);
	return (index ? index->value : def);
}

void ContiguityMatrix::SetByTree(WORD x,WORD y,bool v)
{
	if(!tree) tree=new(node), tree->x=x,tree->y=y,tree->value=v,tree->next1=tree->next2=0,nodeCount=0;
	else ((node*)FindByTree(x,y,true))->value=v;
	if(nodeCount<nodeLimit) return;
	TRACE("***** Restruct matrix\n");
	MakeMatrix();
	TreeToMatrix(tree);
	RemoveTree(tree); tree=0;
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////

NumberMatrix::NumberMatrix()
{
	matrix=0;
	tree=0;
}

NumberMatrix::NumberMatrix(WORD size,BYTE n)
{
	matrix=0;
	tree=0;
	Install(size,n);
}

NumberMatrix::~NumberMatrix()
{
	Uninstall();
}

void NumberMatrix::Install(WORD size_,BYTE n) 
{
	size=size_; nodeLimit=size*(size-1)/2*sizeof(WORD)/sizeof(*tree)+2;
	Uninstall();
	def=n; def<<=8; def+=n;
}

void NumberMatrix::Uninstall()
{
	if(matrix) delete[]matrix;
	if(tree) RemoveTree(tree);
}

//----------------------------------------------------------------

int NumberMatrix::GetSize()
{
	return size;
}

void NumberMatrix::Set(CPoint point,WORD n)
{
	Set((WORD)point.x,(WORD)point.y,n);
}

void NumberMatrix::Set(WORD x,WORD y,WORD v)
{
	if(CheckIndex(x,y,size)) return;
	if(x==y) return; WORD t; if(y<x) t=x,x=y,y=t;
	if (matrix) SetByMatrix(x,y,v); else SetByTree(x,y,v);
}

WORD NumberMatrix::Get(WORD x, WORD y)
{
	if(CheckIndex(x,y,size)) return def;
	if(x==y) return def; WORD t; if(y<x) t=x,x=y,y=t;
//	TRACE("get %d %d  vv=%d\n",x,y,(int)(matrix ? GetByMatrix(x,y) : GetByTree(x,y)));
	return (matrix ? GetByMatrix(x,y) : GetByTree(x,y));
}

//----------------------------------------------------------------
// Matrix as Triangle (c) Proteus

WORD NumberMatrix::GetByMatrix(WORD x,WORD y)
{
	return matrix[IndexByMatrix(x,y)];
}

void NumberMatrix::SetByMatrix(WORD x,WORD y,WORD v)
{
	matrix[IndexByMatrix(x,y)]=v;
}

void NumberMatrix::MakeMatrix()
{
	matrix=new WORD[size*(size-1)/2];
	memset(matrix,def,size*(size-1)/2*sizeof(WORD));
}

//----------------------------------------------------------------
// Matrix as BSP tree  (c) Proteus

void NumberMatrix::TreeToMatrix(node *tree)
{
	if(tree)
	{
		SetByMatrix(tree->x,tree->y,tree->value);
		TreeToMatrix(tree->next1);
		TreeToMatrix(tree->next2);
	}
}

WORD NumberMatrix::GetByTree(WORD x,WORD y)
{
	if(!tree) return def;
	node *index=(node*)FindByTree(x,y);
	return (index ? index->value : def);
}

void NumberMatrix::SetByTree(WORD x,WORD y,WORD v)
{
	if(!tree) tree=new(node), tree->x=x,tree->y=y,tree->value=v,tree->next1=tree->next2=0,nodeCount=0;
	else ((node*)FindByTree(x,y,true))->value=v;
	if(nodeCount<nodeLimit) return;
	MakeMatrix();
	TreeToMatrix(tree);
	RemoveTree(tree); tree=0;
}

void *NumberMatrix::FindByTree(WORD x,WORD y,bool add)
{
	node *next,*root=tree; bool direction,split=false;
	WORD s,sx,sy; s=sx=sy=size/2;
	for(;;)
	{
		if(root->x==x && root->y==y) return root;
		direction=(split?(direction=root->x>sx):(direction=root->y>sy));
		s/=2;
		if(split) {	if(direction) sx+=s; else sx-=s;} else {if(direction) sy+=s; else sy-=s;}
		next=(direction ? root->next1 : root->next2);
		if(next) root=next; else {
			if(!add) return 0; next=new(node); next->next1=next->next2=0; next->x=x; next->y=y; nodeCount++;
			if(direction) root->next1=next; else root->next2=next;
			return next;
		}
	}
}
