// openGL.cpp : implementation file
//

#include "stdafx.h"
#include "Color.h"
#include "openGL.h"
#include "math.h"
#include "OpenglSetup.h"

#include "glut.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// openGL dialog


openGL::openGL(Graph *gr,CWnd* pParent /*=NULL*/)
	: CDialog(openGL::IDD, pParent)
{
	isAnimate=true;

	graph=gr;
	m_hGLContext = NULL;
	m_GLPixelIndex = 0;

	m_xOffset=m_yOffset=0;
	m_xRotate = m_yRotate= m_zRotate = 0;
	m_xtRotate = m_ytRotate = 0;
	LastPoint.x=LastPoint.y=0;
	m_Scale = 1; MouseMode=rotate;
	CurrentTop=-1;
	//{{AFX_DATA_INIT(openGL)
		// NOTE: the ClassWizard will add member initialization here
	//}}AFX_DATA_INIT
}


void openGL::DoDataExchange(CDataExchange* pDX)
{
	CDialog::DoDataExchange(pDX);
	//{{AFX_DATA_MAP(openGL)
		// NOTE: the ClassWizard will add DDX and DDV calls here
	//}}AFX_DATA_MAP
}


BEGIN_MESSAGE_MAP(openGL, CDialog)
	//{{AFX_MSG_MAP(openGL)
	ON_WM_PAINT()
	ON_WM_NCPAINT()
	ON_WM_CREATE()
	ON_WM_TIMER()
	ON_WM_SIZE()
	ON_WM_CONTEXTMENU()
	ON_WM_NCLBUTTONDOWN()
	ON_WM_NCLBUTTONDBLCLK()
	ON_COMMAND(ID_ANIMATE, OnAnimate)
	ON_COMMAND(ID_GLSEL1, OnSelect1)
	ON_COMMAND(ID_GLSEL2, OnSelect2)
	ON_COMMAND(ID_GLSEL3, OnSelect3)
	ON_COMMAND(ID_GLSEL4, OnSelect4)
	ON_UPDATE_COMMAND_UI(ID_ANIMATE, OnUpdateAnimate)
	ON_UPDATE_COMMAND_UI(ID_GLSEL1, OnUpdateSel1)
	ON_UPDATE_COMMAND_UI(ID_GLSEL2, OnUpdateSel2)
	ON_UPDATE_COMMAND_UI(ID_GLSEL3, OnUpdateSel3)
	ON_UPDATE_COMMAND_UI(ID_GLSEL4, OnUpdateSel4)
	ON_COMMAND(ID_SETUP, OnSetup)
	ON_COMMAND(ID_RESET_XY, OnResetXy)
	ON_COMMAND(ID_RESET_SELECT, OnResetSelect)
	ON_COMMAND(ID_RESET_SCALY, OnResetScaly)
	ON_COMMAND(ID_RESET_ANGLE, OnResetAngle)
	ON_COMMAND(ID_MOUSE_MOVE, OnMouseMove1)
	ON_UPDATE_COMMAND_UI(ID_MOUSE_MOVE, OnUpdateMouseMove)
	ON_COMMAND(ID_MOUSE_ROTATE, OnMouseRotate)
	ON_UPDATE_COMMAND_UI(ID_MOUSE_ROTATE, OnUpdateMouseRotate)
	ON_WM_MOUSEMOVE()
	ON_WM_LBUTTONDOWN()
	ON_COMMAND(ID_REDUCE, OnReduce)
	ON_COMMAND(ID_NEXT, OnNext)
	ON_COMMAND(ID_MAGNIFY, OnMagnify)
	ON_COMMAND(ID_BACK, OnBack)
	ON_COMMAND(ID_NORMAL_SIZE, OnNormalSize)
	ON_UPDATE_COMMAND_UI(ID_NORMAL_SIZE, OnUpdateNormalSize)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// openGL message handlers


bool openGL::SetWindowPixelFormat(HDC hDC)
{
	PIXELFORMATDESCRIPTOR pixelDesc;
	pixelDesc.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pixelDesc.nVersion = 1;
	pixelDesc.dwFlags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER | PFD_STEREO_DONTCARE;
	pixelDesc.iPixelType = PFD_TYPE_RGBA;
	pixelDesc.cColorBits = 32;
	pixelDesc.cRedBits = 8;
	pixelDesc.cRedShift = 16;
	pixelDesc.cGreenBits = 8;
	pixelDesc.cGreenShift = 8;
	pixelDesc.cBlueBits = 8;
	pixelDesc.cBlueShift = 0;
	pixelDesc.cAlphaBits = 0;
	pixelDesc.cAlphaShift = 0;
	pixelDesc.cAccumBits = 64;
	pixelDesc.cAccumRedBits = 16;
	pixelDesc.cAccumGreenBits = 16;
	pixelDesc.cAccumBlueBits = 16;
	pixelDesc.cAccumAlphaBits = 0;
	pixelDesc.cDepthBits = 32;
	pixelDesc.cStencilBits = 8;
	pixelDesc.cAuxBuffers = 0;
	pixelDesc.iLayerType = PFD_MAIN_PLANE;
	pixelDesc.bReserved = 0;
	pixelDesc.dwLayerMask = 0;
	pixelDesc.dwVisibleMask = 0;
	pixelDesc.dwDamageMask = 0;

	m_GLPixelIndex = ChoosePixelFormat(hDC,&pixelDesc);
	if(!m_GLPixelIndex) 
	{
		m_GLPixelIndex = 1;
		if(!DescribePixelFormat(hDC,m_GLPixelIndex,sizeof(PIXELFORMATDESCRIPTOR),&pixelDesc))
		return FALSE;
	}
	return SetPixelFormat(hDC,m_GLPixelIndex,&pixelDesc)!=0;
}


bool openGL::CreateViewGLContext(HDC hDC)
{
	m_hGLContext = wglCreateContext(hDC);
	if(!m_hGLContext) return FALSE;
	return wglMakeCurrent(hDC,m_hGLContext)!=0;
}


void openGL::BuildList()
{
	::glNewList(1,GL_COMPILE_AND_EXECUTE);
	CRect rect=graph->GetSize(); center.x=(rect.left+rect.right)/2; center.y=(rect.bottom+rect.top)/2;
	m_Scale_=10.0f/(float)(rect.right-rect.left+rect.bottom-rect.top);
	glPushMatrix(); 
	CPoint link,point1,point2; int cr=setup.LinkSize;
    int n,m,i; double h=sqrt(setup.BollSize*setup.BollSize-cr*cr);
	GLUquadricObj *q;
	GLfloat	diffuseProperties[4]; 
	diffuseProperties[3]=1;
	for(n=0; n<graph->GetLINKs(); n++)
	{
		m=graph->GetColor(n); if(setup.FixColor) m=setup.LinkColor;
		for(i=0;i<3;i++) diffuseProperties[i]=(float)(m&255)/255,m>>=8;
		glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,diffuseProperties);
		q = gluNewQuadric();
		glPushMatrix();
		link=graph->GetLINK(n);
		int dz=graph->GetZ(link.x)-graph->GetZ(link.y);
		point1=graph->GetTOP(link.x);
		point2=graph->GetTOP(link.y)-point1;
		int r=sqrt(point2.x*point2.x+point2.y*point2.y+dz*dz);
		double f1=atan2(-point2.y,-point2.x);
		double f2=asin((double)dz/r);
		glTranslated(point1.x-center.x-cos(f1)*h*cos(f2),point1.y-center.y-sin(f1)*h*cos(f2),graph->GetZ(link.x)-sin(f2)*h);
//-(int)h*2;
		glRotated(f1/PI*180,0.0,0.0,1.0);
		glRotated(-90-f2/PI*180,0.0,1.0,0.0);
		gluCylinder(q,cr,cr,r-h*2,16,1);
//		glColor3d(0,1,0);
//		auxSolidCylinder(cr,r);
//		auxSolidSphere( 9 );
		glPopMatrix();
		gluDeleteQuadric(q);
	}

	diffuseProperties[3]=0.7f; m=setup.BollColor;
	for(i=0;i<3;i++) diffuseProperties[i]=(float)(m&255)/255,m>>=8;
	glMaterialfv(GL_FRONT_AND_BACK,GL_DIFFUSE,diffuseProperties);
	q = gluNewQuadric();
	for(n=0; n<graph->GetTOPs(); n++)
	{
		glPushMatrix();
		point1=graph->GetTOP(n)-center;
		glTranslated(point1.x,point1.y,graph->GetZ(n));
		gluSphere(q,setup.BollSize,16,16);
		glPopMatrix();
	}
	gluDeleteQuadric(q);
	glPopMatrix();
	::glEndList();
}


void openGL::OnPaint() 
{
	CPaintDC dc(this); // device context for painting
	RenderScene();
	SwapBuffers(dc.m_ps.hdc);
//	CDialog::OnPaint();
}


void openGL::RenderScene()
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glPushMatrix();
	glTranslated(m_xOffset,m_yOffset,-8.0);
	
//    float pos[4] = {9,9,9,1};
 //   float dir[3] = {-9,-9,-9};
//	glTranslated(0.0,0.0,-8.0);
 //   glLightfv(GL_LIGHT0, GL_POSITION, pos);
   // glLightfv(GL_LIGHT0, GL_SPOT_DIRECTION, dir);


//    GLfloat mat_specular[] = {1,1,1,1};

 //   glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
 //   glMaterialf(GL_FRONT, GL_SHININESS, 128.0);


	GLUquadricObj *q; double r;
	glScalef(m_Scale*m_Scale_,m_Scale*m_Scale_,m_Scale*m_Scale_);
    // Only one list, the box
	glRotated(m_xRotate,1.0,0.0,0.0);
	glRotated(m_yRotate,0.0,1.0,0.0);
	glRotated(m_zRotate,0.0,0.0,1.0);
	::glCallList(1);
	if(CurrentTop>=0) {
		CPoint point=graph->GetTOP(CurrentTop)-center;
		glTranslated(point.x,point.y,graph->GetZ(CurrentTop));
		glRotated(m_xtRotate,0.0,0.0,1.0);
		glRotated(m_ytRotate,0.0,1.0,0.0);
		switch(setup.SelectType){
		case Torus:	glutSolidTorus(3,setup.BollSize+5,8,20); break;
		case Skeleton: glutSolidTorus(1,setup.BollSize+2,8,20); glRotated(90,0.0,1.0,0.0);
			glutSolidTorus(1,setup.BollSize+2,8,20); glRotated(90,1.0,0.0,0.0);
			glutSolidTorus(1,setup.BollSize+2,8,20); break;
		case Spheres:
			r=setup.BollSize+2;
			q = gluNewQuadric();
			glTranslated(0,0,r);  gluSphere(q,2,10,10);
			glTranslated(0,r,-r);  gluSphere(q,2,10,10);
			glTranslated(r,-r,0);  gluSphere(q,2,10,10);
			glTranslated(-r,-r,0);  gluSphere(q,2,10,10);
			glTranslated(0,r,-r);  gluSphere(q,2,10,10);
			glTranslated(-r,0,r);  gluSphere(q,2,10,10);

			gluDeleteQuadric(q); break;
		case Thorns:
			r=sqrt(setup.BollSize*setup.BollSize-4);
			glTranslated(0,0,r);  glutSolidCone(2,5,10,10);
			glTranslated(0,r,-r); glRotated(-90,1.0,0,0);  glutSolidCone(2,5,10,10);
			glTranslated(0,r,-r); glRotated(-90,1.0,0,0); glutSolidCone(2,5,10,10);
			glTranslated(r,0,-r); glRotated(90,0,1.0,0); glutSolidCone(2,5,10,10);
			glTranslated(0,r,-r); glRotated(-90,1.0,0,0);glutSolidCone(2,5,10,10);
			glTranslated(0,r,-r); glRotated(-90,1.0,0,0);  glutSolidCone(2,5,10,10);
			break;
		}
	}
	glPopMatrix();
}

int openGL::OnCreate(LPCREATESTRUCT lpCreateStruct) 
{
	if (CDialog::OnCreate(lpCreateStruct) == -1) return -1;

	// open GL
	
	HWND hWnd = GetSafeHwnd();  HDC hDC = ::GetDC(hWnd);

	if(!SetWindowPixelFormat(hDC) || !CreateViewGLContext(hDC))	return false;

	// BackColor
//	COLORREF color = ::GetSysColor(COLOR_3DFACE);
//	glClearColor((float)GetRValue(color)/255.0f,(float)GetGValue(color)/255.0f,(float)GetBValue(color)/255.0f,1.0);

	// Lights, material properties
    GLfloat	ambientProperties[]  = {0.7f, 0.7f, 0.7f, 1.0f};
	GLfloat	diffuseProperties[]  = {0.8f, 0.8f, 0.8f, 1.0f};
    GLfloat	specularProperties[] = {1.0f, 1.0f, 1.0f, 1.0f};
	
  glLightfv( GL_LIGHT0, GL_AMBIENT, ambientProperties);
  glLightfv( GL_LIGHT0, GL_DIFFUSE, diffuseProperties);
  glLightfv( GL_LIGHT0, GL_SPECULAR, specularProperties);
  glLightModelf(GL_LIGHT_MODEL_TWO_SIDE, 1.0);

	// Default : lighting
	glEnable(GL_LIGHT0);
	glEnable(GL_LIGHTING);
	glEnable(GL_DEPTH_TEST);
	BuildList();


	SetTimer(0,10,NULL);
	return 0;
}

void openGL::OnTimer(UINT nIDEvent) 
{
	if(isAnimate) {
		m_xRotate -= 1; if (m_xRotate<0) m_xRotate+=360;
		m_zRotate -= 2; if (m_zRotate<0) m_zRotate+=360;
	}
	m_xtRotate -= 3; if (m_xtRotate<0) m_xtRotate+=360;
	m_ytRotate -= 5; if (m_ytRotate<0) m_ytRotate+=360;

//	CRect rect(0,30,200,200);
	InvalidateRect(0,FALSE);
	CDialog::OnTimer(nIDEvent);
}

void openGL::OnSize(UINT nType, int cx, int cy) 
{
	CDialog::OnSize(nType, cx, cy);
	
	GLsizei width,height;
	GLdouble aspect;

	width = cx;
	height = cy;

	if(cy==0)
		aspect = (GLdouble)width;
	else
		aspect = (GLdouble)width/(GLdouble)height;

	glViewport(0,0,width,height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(45,aspect,1,20.0);
    glMaterialf(GL_FRONT, GL_SHININESS, 128);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glEnable(GL_NORMALIZE);

}


void openGL::OnNcPaint()
{
	Default();
	CDC* pDC = GetWindowDC();
	CRect rc; int limit;
	GetWindowRect(rc);
	limit=rc.right-GetSystemMetrics(SM_CYBORDER);
	int x=GetSystemMetrics(SM_CXFRAME);
	int y=GetSystemMetrics(SM_CYFRAME);
	int l=GetSystemMetrics(SM_CXSIZE);
	int h=GetSystemMetrics(SM_CYSIZE);

	for(int n=0; n<8; n++) {
		x+=l+l/2;
		if(x+l>limit) break;
		DrawIconEx(pDC->m_hDC,x,y,AfxGetApp()->LoadIcon(IDI_GL1+n),l,h,0,0,DI_NORMAL);
	}
}

void openGL::OnContextMenu(CWnd* pWnd, CPoint point) 
{
   CMenu menu, *pSubMenu; if(!menu.LoadMenu(IDR_OPENGL)) return;
   pSubMenu = menu.GetSubMenu(0); if (!pSubMenu) return;
   UpdateMenu(this,pSubMenu);
   ::TrackPopupMenu(pSubMenu->m_hMenu, 0,point.x,point.y, 0, pWnd->GetSafeHwnd(), NULL);
   Invalidate();
}

void openGL::OnNcLButtonDown(UINT nHitTest, CPoint point)
{
	OnNcClick(nHitTest,point,false);
}


void openGL::OnNcLButtonDblClk(UINT nHitTest, CPoint point)
{
	OnNcClick(nHitTest,point,true);
}


void openGL::OnNcClick(UINT nHitTest, CPoint point,bool doub)
{
	CRect rect; CPoint point1=point;
	ScreenToClient(&point);  CMenu menu, *pSubMenu; 
	int n=GetSystemMetrics(SM_CYFRAME);
	if(-point.y>=n && -point.y<=GetSystemMetrics(SM_CYSIZE)+n) {
	int l=GetSystemMetrics(SM_CXSIZE); int x=0;
	for(n=0; n<8; n++)
	{
		x+=l+l/2;
		if(point.x>=x && point.x<=x+l) switch(n) {
			case 0:
			case 5:
				Sound(IDW_PIP);
				GetWindowRect(&rect);
				if(!menu.LoadMenu(IDR_OPENGL)) return;
				pSubMenu = menu.GetSubMenu(n ? 2 : 1); 
				if (!pSubMenu) return;
				UpdateMenu(this,pSubMenu);
				::TrackPopupMenu(pSubMenu->m_hMenu, 0,rect.left+x,rect.top+GetSystemMetrics(SM_CYSIZE)+GetSystemMetrics(SM_CYFRAME), 0, this->GetSafeHwnd(), NULL); return;
			case 1: Sound(IDW_PIP); CurrentTop=(CurrentTop+1)%graph->GetTOPs(); return;
			case 2: Sound(IDW_PIP); n=graph->GetTOPs(); CurrentTop=(CurrentTop+n-1)%n; return;
			case 3: if(CurrentTop>=0) graph->SetZ(CurrentTop,graph->GetZ(CurrentTop)+1),BuildList(); return;
			case 4: if(CurrentTop>=0) graph->SetZ(CurrentTop,graph->GetZ(CurrentTop)-1),BuildList(); return;
			case 6: m_Scale=m_Scale*1.05f; return;
			case 7: m_Scale=m_Scale*0.95f; return;
		}
	}}
	if (doub) CDialog::OnNcLButtonDblClk(nHitTest,point1);
	else CDialog::OnNcLButtonDown(nHitTest,point1);
}


void openGL::OnAnimate() 
{
	isAnimate=!isAnimate;
}

void openGL::OnUpdateAnimate(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(isAnimate);
}

void openGL::OnUpdateSel1(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(setup.SelectType==Torus);
}

void openGL::OnUpdateSel2(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(setup.SelectType==Skeleton);
}

void openGL::OnUpdateSel3(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(setup.SelectType==Spheres);
}

void openGL::OnUpdateSel4(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(setup.SelectType==Thorns);
}

void openGL::OnSelect1()
{
	setup.SelectType=Torus;
}
void openGL::OnSelect2()
{
	setup.SelectType=Skeleton;
}
void openGL::OnSelect3()
{
	setup.SelectType=Spheres;
}
void openGL::OnSelect4()
{
	setup.SelectType=Thorns;
}

void openGL::OnSetup() 
{
	OpenglSetup dlg;
	dlg.color1=setup.BollColor;
	dlg.color2=setup.LinkColor;
	dlg.m_size1=setup.BollSize;
	dlg.m_size2=setup.LinkSize;
	dlg.m_fix=setup.FixColor;
	dlg.m_mode=setup.SelectType;
	dlg.m_magnify=m_Scale*100;
	dlg.m_x=m_xOffset*100;
	dlg.m_y=m_yOffset*100;
	if(dlg.DoModal()!=IDOK) return;
	m_xOffset=(double)dlg.m_x/100;
	m_yOffset=(double)dlg.m_y/100;
	m_Scale=(double)dlg.m_magnify/100;
	setup.SelectType=(selectE)dlg.m_mode;
	setup.FixColor=dlg.m_fix!=0;
	setup.LinkSize=dlg.m_size2;
	setup.BollSize=dlg.m_size1;
	if(setup.LinkSize>setup.BollSize) setup.LinkSize=setup.BollSize=(setup.LinkSize+setup.BollSize)/2;
	setup.BollColor=dlg.color1;
	setup.LinkColor=dlg.color2;
	BuildList();
}

void openGL::OnResetXy() 
{
	m_xOffset=m_yOffset=0;	
}

void openGL::OnResetSelect() 
{
	CurrentTop=-1;	
}

void openGL::OnResetScaly() 
{
	m_Scale=1;	
}

void openGL::OnResetAngle() 
{
	m_xRotate=m_yRotate=m_zRotate=0;
}

void openGL::OnMouseMove1() 
{
	MouseMode=move;
}

void openGL::OnUpdateMouseMove(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(MouseMode==move);
}

void openGL::OnMouseRotate() 
{
	MouseMode=rotate;
}

void openGL::OnUpdateMouseRotate(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(MouseMode==rotate);
}

void openGL::OnMouseMove(UINT nFlags, CPoint point) 
{
	if(!(nFlags & MK_LBUTTON)) return;
	CPoint temp=point-pullPoint;
	if(MouseMode==move) 
	{
		m_xOffset+=(double)temp.x/20;
		m_yOffset-=(double)temp.y/20;
	} else {
		m_xRotate+=(double)temp.y/20;
		m_zRotate+=(double)temp.x/20;
	}
	pullPoint=point;
}

void openGL::OnLButtonDown(UINT nFlags, CPoint point) 
{
	pullPoint=point;
}

void openGL::OnReduce() 
{
   m_Scale=m_Scale*0.9f;	
}

void openGL::OnNext() 
{
	CurrentTop=(CurrentTop+1)%graph->GetTOPs();
}

void openGL::OnMagnify() 
{
	m_Scale=m_Scale*1.1f;
}

void openGL::OnBack() 
{
	int n=graph->GetTOPs(); CurrentTop=(CurrentTop+n-1)%n;
}

void openGL::OnNormalSize() 
{
   m_Scale=1;	
}

void openGL::OnUpdateNormalSize(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_Scale==1);	
}
